// AY-8- Driver
// (c) 2016 Eleken
// rev_c1815@yahoo.co.jp

#include <xc.h>

#include "psg.h"
#include "sreg.h"

// SY-3-8910 Control
// Note No -> Divider (main clock = 1.25 MHz)
uint16_t g_divTable[128] = {
	4095, 4095, 4095, 4095, 4095, 4095, 4095, 4095, 4095, 4095, 4095, 4095, //0-11
	4095, 4095, 4095, 4095, 4095, 4095, 4054, 3827, 3612, 3409, 3218, 3037, //12-23
	2867, 2706, 2554, 2411, 2275, 2148, 2027, 1913, 1806, 1705, 1609, 1519, //24-35
	1433, 1353, 1277, 1205, 1138, 1074, 1014, 957, 903, 852, 804, 759, //36-47
	717, 676, 638, 603, 569, 537, 507, 478, 451, 426, 402, 380, //48-59
	358, 338, 319, 301, 284, 268, 253, 239, 226, 213, 201, 190, //60-71
	179, 169, 160, 151, 142, 134, 127, 120, 113, 107, 101, 95, //72-
	90, 85, 80, 75, 71, 67, 63, 60, 56, 53, 50, 47, 
	45, 42, 40, 38, 36, 34, 32, 30, 28, 27, 25, 24, 
	22, 21, 20, 19, 18, 17, 16, 15, 14, 13, 13, 12, 
	11, 11, 10, 9, 9, 8, 8, 7
};

void SetRegister(uint8_t addr, uint8_t value)
{
	SregSend(addr);
	//__delay_us(1);
	PORTC = (PORTC & 0x3F) | (3 << 6); // ALE
	//asm("nop");
	__delay_us(1);
	PORTC = (PORTC & 0x3F) | 0x00; // NOP
	asm("nop");
	//__delay_us(1);

	SregSend(value);
	__delay_us(1);
	PORTC = (PORTC & 0x3F) | (1 << 6); // WRITE
	//asm("nop");
	__delay_us(1);
	PORTC = (PORTC & 0x3F) | 0x00; // NOP
	//__delay_us(1);
}



uint8_t g_amplitudeState[3]; // 0-15
uint8_t g_noteVelocityState[3]; // 0-127
uint8_t g_envelope;
uint8_t g_enable;
uint8_t g_lastNoteNo[3];
int16_t g_lastPitchBend[3];

static uint16_t GetDividerValue(uint8_t ch)
{
	uint16_t temp = (g_lastPitchBend[ch] + 8208);
	uint8_t note = g_lastNoteNo[ch] + temp/342 - 24;
	uint8_t w = ((uint32_t)(temp%342) )*256/342;
	
	//uint8_t note = g_lastNoteNo[ch];
	//uint8_t w = 0;
	
	uint32_t div1 = g_divTable[note];
	uint32_t div2 = g_divTable[note+1];
	uint16_t div = (div1 * (256-w) + div2 * w)/256;
	
	return div;
}

void InitPSG(void)
{
	SregInit();
	
	SetRegister(0x06, 0x01);
	g_enable = 0b00111111;
	SetRegister(0x07, g_enable);
	SetRegister(0x08, 0x0f); // Constant mode, ampl = max
	SetRegister(0x09, 0x0f);
	SetRegister(0x0A, 0x0f);

	SetRegister(0x0D, 0x00);
	
	SendNoteMode(0, 0); // Constant Mode
	SendNoteMode(1, 0);
	SendNoteMode(2, 0);
	
	SendNoteVol(0, 15); // 0-15
	SendNoteVol(1, 15); // 0-15
	SendNoteVol(2, 15); // 0-15
	
	g_lastPitchBend[0] = 0;
	g_lastPitchBend[1] = 0;
	g_lastPitchBend[2] = 0;
	
	SregSend(0x3F);
}

void SendNoteOn(uint8_t ch, uint8_t note)
{
	g_lastNoteNo[ch] = note;
	
	// Tone
	uint16_t div = GetDividerValue(ch);
	
	SetRegister(ch*2, div);
	SetRegister(ch*2+1, div>>8);
	
	
	// Amp
	//g_noteVelocityState[ch] = velocity;
	//SendNoteVol(g_amplitudeState[ch]);
	
	// Enable
	SendToneEnable(ch, 1);

	// Envelope
	if(g_amplitudeState[ch] & 0x10) {
		SetRegister(13, g_envelope); // Write envelope to clear internal counter
	}
	
	// test.
	//SetRegister(ch*2+1, div>>8);
	//SetRegister(ch*2+1, div>>8);
	
	if(ch <= 2){
		SBI(PORTC, ch); // LED
	}
}
void SendNoteOff(uint8_t ch)
{
//	SetRegister(ch+8, 0);

	SendToneEnable(ch, 0);
	
	if(ch <= 2){
		CBI(PORTC, ch); // LED
	}
}
void SendNoteVol(uint8_t ch, uint8_t vol)
{
	uint8_t newValue = (g_amplitudeState[ch] & 0xF0) | vol;
	
	if(g_amplitudeState[ch] == newValue) return;
	
	g_amplitudeState[ch] = (g_amplitudeState[ch] & 0xF0) | vol;
	SetRegister(ch+8, g_amplitudeState[ch]);
}
void SendNoteMode(uint8_t ch, uint8_t mode) // mode=1: envelope, mode=0: constant volume
{
	uint8_t newValue = (mode << 4) | (g_amplitudeState[ch] & 0x0F);
	
	if(g_amplitudeState[ch] == newValue) return;
	
	g_amplitudeState[ch] = newValue;
	SetRegister(ch+8, g_amplitudeState[ch]);
}
void SendToneEnable(uint8_t ch, uint8_t enableState)
{
	
	if(ch == 0){
		g_enable = (g_enable & 0b11111110);
		if(!(enableState & 0x01)) g_enable |= 0b00000001;
	}else if(ch == 1){
		g_enable = (g_enable & 0b11111101);
		if(!(enableState & 0x01)) g_enable |= 0b00000010;
	}else{
		g_enable = (g_enable & 0b11111011);
		if(!(enableState & 0x01)) g_enable |= 0b00000100;
	}

	SetRegister(7, g_enable);
}
void SendNoiseEnable(uint8_t ch, uint8_t enableState, uint8_t vol)
{
	// note that value 0 of enable register means enabled
	if(ch == 0){
		g_enable = (g_enable & 0b11110111);
		if(!(enableState & 0x01)) g_enable |= 0b00001000;
	}else if(ch == 1){
		g_enable = (g_enable & 0b11101111);
		if(!(enableState & 0x01)) g_enable |= 0b00010000;
	}else{
		g_enable = (g_enable & 0b11011111);
		if(!(enableState & 0x01)) g_enable |= 0b00100000;
	}
	//SetRegister(7, g_enable);
	// [TEST] when noise is enabled, clear tone enable
	{
		uint8_t enable = g_enable;
		if(!(enable & 0b00001000)) enable |= 0x01;
		if(!(enable & 0b00010000)) enable |= 0x02;
		if(!(enable & 0b00100000)) enable |= 0x04;
		
		SetRegister(7, enable);
		
	}
	
	if(enableState & 0x01){ // set vol
		SetRegister(ch+8, (g_amplitudeState[ch] & 0xF0) | vol);
	}else{ // restore vol
		SetRegister(ch+8, g_amplitudeState[ch]);
	}
}
void SendNoisePeriod(uint8_t value)
{
	SetRegister(6, value);
}
void SendEnvelopePeriod(uint16_t value)
{
	SetRegister(11, value);
	SetRegister(12, value >> 8);
}
void SendEnvelopeType(uint8_t value)
{
	g_envelope = value;
	SetRegister(13, g_envelope);
}
void SendPitchBend(uint8_t ch, int16_t bend)
{
	g_lastPitchBend[ch] = bend;
	
	// Tone
	uint16_t div = GetDividerValue(ch);
	
	SetRegister(ch*2, div);
	SetRegister(ch*2+1, div>>8);
}

